
from pathlib import Path
from typing import Optional, Union, TypeVar, Type, Literal, Any, Callable, Iterable, Hashable, ParamSpec

from crocodile.core_modules.core_2 import Base, List
from crocodile.core_modules.core_4 import Display


T = TypeVar('T')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
PLike = Union[str, Path]
PS = ParamSpec('PS')


class Struct(Base):  # inheriting from dict gives `get` method, should give `__contains__` but not working. # Inheriting from Base gives `save` method.
    """Use this class to keep bits and sundry items. Combines the power of dot notation in classes with strings in dictionaries to provide Pandas-like experience"""
    def __init__(self, dictionary: Union[dict[Any, Any], Type[object], None] = None, **kwargs: Any):
        if dictionary is None or isinstance(dictionary, dict): final_dict: dict[str, Any] = {} if dictionary is None else dictionary
        else:
            final_dict = (dict(dictionary) if dictionary.__class__.__name__ == "mappingproxy" else dictionary.__dict__)  # type: ignore
        final_dict.update(kwargs)  # type ignore
        super(Struct, self).__init__()
        self.__dict__ = final_dict  # type: ignore
    @staticmethod
    def from_keys_values(k: Iterable[str], v: Iterable[Any]) -> 'Struct': return Struct(dict(zip(k, v)))
    @staticmethod
    def from_keys_values_pairs(my_list: list[tuple[Any, Any]]) -> "Struct": return Struct({k: v for k, v in my_list})
    @staticmethod
    def from_names(names: list[str], default_: Optional[Any] = None) -> 'Struct': return Struct.from_keys_values(k=names, v=default_ or [None] * len(names))  # Mimick NamedTuple and defaultdict
    def spawn_from_values(self, values: Union[list[Any], List[Any]]) -> 'Struct': return self.from_keys_values(list(self.keys()), values)
    def spawn_from_keys(self, keys: Union[list[str], List[str]]) -> 'Struct': return self.from_keys_values(keys, list(self.values()))
    def to_default(self, default: Optional[Callable[[], Any]] = lambda: None):
        import collections
        tmp2: dict = collections.defaultdict(default)  # type: ignore
        tmp2.update(self.__dict__)
        self.__dict__ = tmp2
        return self
    def __str__(self, sep: str = "\n"): return Display.config(self.__dict__, sep=sep)
    def __getattr__(self, item: str) -> 'Struct':
        try: return self.__dict__[item]
        except KeyError as ke: raise AttributeError(f'{type(self).__name__!r} object has no attribute {item!r}') from ke  # this works better with the linter. replacing Key error with Attribute error makes class work nicely with hasattr() by returning False.
    # clean_view = property(lambda self: type("TempClass", (object,), self.__dict__))
    def __repr__(self, limit: int = 150): return "Struct: " + Display.get_repr(self.keys().list.__repr__(), limit=limit, justify=0)
    def __getitem__(self, item: str): return self.__dict__[item]  # thus, gives both dot notation and string access to elements.
    def __setitem__(self, key: str, value: Any): self.__dict__[key] = value
    def __bool__(self): return bool(self.__dict__)
    def __contains__(self, key: Hashable): return key in self.__dict__
    def __len__(self): return len(self.keys())
    def __getstate__(self): return self.__dict__  # serialization
    def __setstate__(self, state: dict[Any, Any]): self.__dict__ = state
    def __iter__(self): return iter(self.__dict__.items())
    def __delitem__(self, key: str): del self.__dict__[key]
    def copy(self) -> 'Struct': return Struct(self.__dict__.copy())
    def to_dataframe(self, *args: Any, **kwargs: Any):
        import pandas as pd
        return pd.DataFrame(self.__dict__, *args, **kwargs)
    def keys(self, verbose: bool = False) -> 'List[Any]':
        from tqdm import tqdm
        return List(list(self.__dict__.keys()) if not verbose else tqdm(self.__dict__.keys()))
    def values(self, verbose: bool = False) -> 'List[Any]':
        from tqdm import tqdm
        return List(list(self.__dict__.values()) if not verbose else tqdm(self.__dict__.values()))
    def items(self, verbose: bool = False, desc: str = "") -> 'List[Any]':
        from tqdm import tqdm
        return List(self.__dict__.items() if not verbose else tqdm(self.__dict__.items(), desc=desc))
    def get(self, key: Optional[str] = None, default: Optional[Any] = None, strict: bool = False, keys: Union[None, list[str]] = None) -> 'Union[Any, List[Any]]':
        if keys is not None: return List([self.__dict__.get(key, default) if not strict else self[key] for key in keys])
        if key is not None: return (self.__dict__.get(key, default) if not strict else self[key])
        else: raise ValueError("Either key or keys should be passed.")
    def apply2keys(self, kv_func: Callable[[Any, Any], Any], verbose: bool = False, desc: str = "") -> 'Struct': return Struct({kv_func(key, val): val for key, val in self.items(verbose=verbose, desc=desc)})
    def apply2values(self, kv_func: Callable[[Any, Any], Any], verbose: bool = False, desc: str = "") -> 'Struct':
        _ = [self.__setitem__(key, kv_func(key, val)) for key, val in self.items(verbose=verbose, desc=desc)]
        return self
    def apply(self, kv_func: Callable[[Any, Any], Any]) -> 'List[Any]': return self.items().apply(lambda item: kv_func(item[0], item[1]))
    def filter(self, kv_func: Callable[[Any, Any], Any]) -> 'Struct': return Struct({key: self[key] for key, val in self.items() if kv_func(key, val)})
    def inverse(self) -> 'Struct': return Struct({v: k for k, v in self.__dict__.items()})
    def update(self, *args: Any, **kwargs: Any) -> 'Struct': self.__dict__.update(Struct(*args, **kwargs).__dict__); return self
    def delete(self, key: Optional[str] = None, keys: Optional[list[str]] = None, kv_func: Optional[Callable[[Any, Any], Any]] = None) -> 'Struct':
        for key in ([key] if key else [] + (keys if keys is not None else [])): self.__dict__.__delitem__(key)
        if kv_func is not None:
            for k, v in self.items():
                if kv_func(k, v): self.__dict__.__delitem__(k)
        return self
    def _pandas_repr(self, justify: int, return_str: bool = False, limit: int = 30):
        import pandas as pd
        import numpy as np
        col2: List[Any] = self.values().apply(lambda x: str(type(x)).split("'")[1])
        col3: List[Any] = self.values().apply(lambda x: Display.get_repr(x, justify=justify, limit=limit).replace("\n", " "))
        array = np.array([self.keys(), col2, col3]).T
        res: pd.DataFrame = pd.DataFrame(array, columns=["key", "dtype", "details"])
        return res if not return_str else str(res)
    def print(self, dtype: bool = True, return_str: bool = False, justify: int = 30, as_config: bool = False, as_yaml: bool = False,  # type: ignore # pylint: disable=W0237
              limit: int = 50, title: str = "", attrs: bool = False, **kwargs: Any) -> Union[str, None]:  # type: ignore
        _ = attrs
        import pandas as pd
        if as_config and not return_str:
            from rich import inspect
            inspect(self, value=False, title=title, docs=False, dunder=False, sort=False)
            return None
        if not bool(self):
            if return_str: return "Empty Struct."
            else: print("Empty Struct."); return None
        else:
            if as_yaml or as_config:
                import yaml
                tmp: str = yaml.dump(self.__dict__) if as_yaml else Display.config(self.__dict__, justify=justify, **kwargs)
                if return_str: return tmp
                else:
                    from rich.syntax import Syntax
                    from rich.console import Console
                    console = Console()
                    console.print(Syntax(tmp, "yaml"))
                    return None
            else:
                tmp2 = self._pandas_repr(justify=justify, return_str=False, limit=limit)
                if isinstance(tmp2, pd.DataFrame):
                    res = tmp2.drop(columns=[] if dtype else ["dtype"])
                else: raise TypeError(f"Unexpected type {type(tmp2)}")
                if not return_str:
                    # import tabulate
                    import rich
                    rich.print(res.to_markdown())
                    # else: print(res)
                    return None
                return str(res)
    @staticmethod
    def concat_values(*dicts: dict[Any, Any], orient: Literal["dict", "list", "series", "split", "tight", "index"] = 'list') -> 'Struct':
        import pandas as pd
        tmp = [Struct(x).to_dataframe() for x in dicts]
        res = pd.concat(tmp).to_dict(orient=orient)
        return Struct(res)  # type: ignore
    def plot_plt(self, title: str = '', xlabel: str = '', ylabel: str = '', **kwargs: Any):
        from crocodile.matplotlib_management import LineArtist
        artist = LineArtist(figname='Structure Plot', **kwargs)
        artist.plot_dict(self.__dict__, title=title, xlabel=xlabel, ylabel=ylabel)
        return artist
    def plot_plotly(self):
        import plotly.express as px
        fig = px.line(self.__dict__)
        fig.show()
        return fig




"""DB
"""

import numpy as np
import pandas as pd
from sqlalchemy import create_engine, text
import psutil

engine = create_engine('sqlite:///:memory:')


def generate_random_dataframe():
    num_rows = 1000000
    num_columns = 10
    column_names: list[str] = [''.join(np.random.choice(list('abcdefghijklmnopqrstuvwxyz'), 5)) for _ in range(num_columns)]
    data = np.random.rand(num_rows, num_columns)
    df_ = pd.DataFrame(data, columns=column_names)
    return df_


for i in range(10):
    df = generate_random_dataframe()
    table_name = 'table_' + str(i)
    df.to_sql(table_name, engine, index=False)


memory_before = psutil.virtual_memory().used / (1024 ** 3)  # Memory usage in GB
print(f"Memory Usage Before Dropping Tables: {memory_before:.2f} GB")

# Loop over the tables, drop them, and check memory usage after each drop
for i in range(10):
    table_name = 'table_' + str(i)
    connection = engine.connect()
    connection.execute(text(f'DROP TABLE IF EXISTS {table_name}'))
    connection.close()

memory_final = psutil.virtual_memory().used / (1024 ** 3)  # Memory usage in GB
print(f"Memory Usage After Dropping All Tables: {memory_final:.2f} GB")

engine.pool.dispose()
memory_final = psutil.virtual_memory().used / (1024 ** 3)  # Memory usage in GB
print(f"Memory Usage After Disposing Connections: {memory_final:.2f} GB")



from crocodile.core import install_n_import, Struct
from crocodile.file_management import P, PLike
import time
from typing import Any, Optional, Callable, Literal, TypeVar, ParamSpec

T = TypeVar('T')
PS = ParamSpec('PS')


def generate_readme(path: PLike, obj: Any = None, desc: str = '', save_source_code: bool = True, verbose: bool = True):  # Generates a readme file to contextualize any binary files by mentioning module, class, method or function used to generate the data"""
    import inspect
    text: str = "# Description\n" + desc + (separator := "\n" + "-" * 50 + "\n\n")
    obj_path = P(inspect.getfile(obj)) if obj is not None else None
    path = P(path)
    if obj_path is not None:
        text += f"# Source code file generated me was located here: \n`{obj_path.collapseuser().as_posix()}`\n" + separator
        try:
            from git import Repo
            repo = Repo(obj_path.parent, search_parent_directories=True)
            text += f"# Last Commit\n{repo.git.execute(['git', 'log', '-1'])}{separator}# Remote Repo\n{repo.git.execute(['git', 'remote', '-v'])}{separator}"
            try: tmppp = obj_path.relative_to(repo.working_dir).as_posix()
            except Exception: tmppp = ""  # type: ignore
            text += f"# link to files: \n{repo.remote().url.replace('.git', '')}/tree/{repo.active_branch.commit.hexsha}/{tmppp}{separator}"
        except Exception as ex: text += f"Could not read git repository @ `{obj_path.parent}`\n{ex}.\n"
    if obj is not None:
        try: source_code = inspect.getsource(obj)
        except OSError: source_code = f"Could not read source code from `{obj_path}`."
        text += ("\n\n# Code to reproduce results\n\n```python\n" + source_code + "\n```" + separator)
    readmepath = (path / "README.md" if path.is_dir() else (path.with_name(path.trunk + "_README.md") if path.is_file() else path)).write_text(text, encoding="utf-8")
    if verbose: print(f"💽 SAVED {readmepath.name} @ {readmepath.absolute().as_uri()}")
    if save_source_code:
        if hasattr(obj, "__code__"):
            save_path = obj.__code__.co_filename
        else:
            module_maybe = inspect.getmodule(obj)
            if module_maybe is not None: save_path = module_maybe.__file__
            else: save_path = None
        if save_path is None:
            print(f"Could not find source code for {obj}.")
            return readmepath
        P(save_path).zip(path=readmepath.with_name(P(readmepath).trunk + "_source_code.zip"), verbose=False)
        print("💽 SAVED source code @ " + readmepath.with_name("source_code.zip").absolute().as_uri())
        return readmepath


class RepeatUntilNoException:
    """
    Repeat function calling if it raised an exception and/or exceeded the timeout, for a maximum of `retry` times.
    * Alternative: `https://github.com/jd/tenacity`
    """
    def __init__(self, retry: int, sleep: float, timeout: Optional[float] = None, scaling: Literal["linear", "exponential"] = "exponential"):
        self.retry = retry
        self.sleep = sleep
        self.timeout = timeout
        self.scaling: Literal["linear", "exponential"] = scaling
    def __call__(self, func: Callable[PS, T]) -> Callable[PS, T]:
        from functools import wraps
        if self.timeout is not None:
            func = install_n_import("wrapt_timeout_decorator").timeout(self.timeout)(func)
        @wraps(wrapped=func)
        def wrapper(*args: PS.args, **kwargs: PS.kwargs):
            t0 = time.time()
            for idx in range(self.retry):
                try:
                    return func(*args, **kwargs)
                except Exception as ex:
                    match self.scaling:
                        case "linear":
                            sleep_time = self.sleep * (idx + 1)
                        case "exponential":
                            sleep_time = self.sleep * (idx + 1)**2
                    print(f"""💥 [RETRY] Function {func.__name__} call failed with error:
{ex}
Retry count: {idx}/{self.retry}. Sleeping for {sleep_time} seconds.
Total elapsed time: {time.time() - t0:0.1f} seconds.""")
                    print(f"""💥 Robust call of `{func}` failed with ```{ex}```.\nretrying {idx}/{self.retry} more times after sleeping for {sleep_time} seconds.\nTotal wait time so far {time.time() - t0: 0.1f} seconds.""")
                    time.sleep(sleep_time)
            raise RuntimeError(f"💥 Robust call failed after {self.retry} retries and total wait time of {time.time() - t0: 0.1f} seconds.\n{func=}\n{args=}\n{kwargs=}")
        return wrapper


def show_globals(scope: dict[str, Any], **kwargs: Any):
    # see print_dir
    return Struct(scope).filter(lambda k, v: "__" not in k and not k.startswith("_") and k not in {"In", "Out", "get_ipython", "quit", "exit", "sys"}).print(**kwargs)
